
// NamedCaptureMixers.java
// Andrew Davison, March 2012, ad@fivedots.coe.psu.ac.th

/* Print out lots of information on the capture mixer whose name contains the string
   supplied on the command line.

   A capture mixer is either:
         1) a TargetDataLine  (for incoming audio):
      or
         2) for a port on a mixer source line (for the recording controls) 

   For each TargetDataLine, all the supported audio formats are listed.
   For an input port, all the supported controls on its source lines are listed.
*/

import java.util.*;
import javax.sound.sampled.*;


public class NamedCaptureMixers
{

  public static void main(String[] args)
  {
    if (args.length != 1) {
      System.out.println("Usage: java NamedCaptureMixers <mixer name>");
      return;
    }

    System.out.println("Looking for capture mixers whose name contains: \"" + args[0] + "\"\n");
    int numCaptures = reportNamedCaptures(args[0]);
    if (numCaptures == 0)
      System.out.println("  None found");
  }  // end of main()



  private static int reportNamedCaptures(String nm)
  /*  An input mixer:
         1) has no source lines and
            no port mixer in the target lines
      or
         2) has a port mixer in a source line

    Note that Ports are separated into source (for recording) and target 
   (for playback). The source/target sense here is from the perspective 
    of the Mixer
  */
  {
    int numCaptures = 0;
    for (Mixer.Info mi : AudioSystem.getMixerInfo()) {
      Mixer mixer = AudioSystem.getMixer(mi);
      if (AudioUtils.containsName(mixer, nm)) {
        if (AudioUtils.isTargetDataLine(mixer)) {
          numCaptures++;
          printTDLInfo(numCaptures, mixer);
        }
        else if (AudioUtils.isInputPort(mixer)) {
          numCaptures++;
          printInputPortInfo(numCaptures, mixer);
        }
      }
    }
    return numCaptures;
  }  // end of reportNamedCaptures()



  private static void printTDLInfo(int idNum, Mixer mixer)
  // For each TargetDataLine, list all the supported audio formats
  {
    AudioUtils.printInfo(idNum, mixer);

    for (Line.Info info : mixer.getTargetLineInfo()) {
      System.out.println("   Target line: " + info);
      DataLine.Info di = (DataLine.Info) info;
      printAudioFormats(di.getFormats());
    }
    System.out.println();
  }  // end of printTDLInfo()



  private static void printAudioFormats(AudioFormat[] formats)
  {
    for (int i = 0; i < formats.length; i++)
      System.out.println("   " + (i+1) + ": " + formats[i]);
    if (formats.length == 0)
      System.out.println("   [no formats]");
    System.out.println();
  }  // end of printAudioFormats()



  private static void printInputPortInfo(int idNum, Mixer mixer)
  /* List all the supported controls on the source lines of the mixer,
     and any target lines */
  {
    AudioUtils.printInfo(idNum, mixer);

    for (Line.Info lineInfo : mixer.getSourceLineInfo()) {
      System.out.println("   Source line: " + lineInfo);
      accessControls(mixer, lineInfo);
    }

    for (Line.Info lineInfo : mixer.getTargetLineInfo()) {
      System.out.println("   Target line: " + lineInfo);
      accessControls(mixer, lineInfo);
    }
    System.out.println();

  }  // end of printInputPortInfo()



  private static void accessControls(Mixer mixer, Line.Info info)
  // access the source port's controls by openning it
  {
    Line line = null;
    try {
      line = mixer.getLine(info);
      line.open();
      printControls(line.getControls());
    }
    catch (LineUnavailableException e) {
      System.out.println("Line could not be openned");
    }
    finally {
      if (line != null)
        line.close();
    }
  }  // end of accessControls()



  private static void printControls(Control[] controls)
  {
    for (int i = 0; i < controls.length; i++)
      printControl("   ", (i+1), controls[i]);
    if (controls.length == 0)
      System.out.println("  [no controls]");
    System.out.println();
  }  // end of printControls()




  private static void printControl(String indent, int id, Control control)
  /* print controls information. A control's type can be
     boolean, compound, enum, or float. Compound and enum controls contain
     other controls, and so printControl() recursively searches them.
  */
  {
    String type = control.getType().toString();

    if (control instanceof BooleanControl) {
      BooleanControl ctrl = (BooleanControl) control;
      System.out.println(indent + id + " Boolean Control: " + ctrl);
    }
    else if (control instanceof CompoundControl) {
      CompoundControl ctrl = (CompoundControl) control;
      Control[] ctrls = ctrl.getMemberControls();
      System.out.println(indent + id + " Compound Control: " + type + " - values below");
      for (int i=0; i < ctrls.length; i++)
        printControl(indent + "  ", (i+1), ctrls[i]);
    }
    else if (control instanceof EnumControl) {
      EnumControl ctrl = (EnumControl) control;
      Object[] values = ctrl.getValues();
      Object value = ctrl.getValue();
      System.out.println(indent + id + " Enum Control: " + type + " - values below");
                 // choose from a set of objects
      for (int i=0; i < values.length; i++) {
        if (values[i] instanceof Control) {
          printControl(indent + "  ", (i+1), (Control) values[i]);
        }
        else
          System.out.println(indent + "  Values[" + (i+1) + "]: " +
                           ((values[i] == value) ? "*" : "") + values[i]);
      }
    }
    else if (control instanceof FloatControl) {
      FloatControl ctrl = (FloatControl) control;
      System.out.println(indent + id + " Float Control: " + ctrl);
    }
    else
      System.out.println(indent + id + " Control: " + control);
  }   // end of printControl()


}  // end of NamedCaptureMixers class

