/* SoundLocalizer - a sound source localization program
 * Copyright (C) 2010 Laurent Calmes
 *
 * This file is part of SoundLocalizer.
 *
 * SoundLocalizer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SoundLocalizer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SoundLocalizer.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.laurentcalmes.soundloc;

import java.io.IOException;

import javax.sound.sampled.AudioFormat;

/**
 * Abstract base class for audio devices
 * 
 * @author Laurent Calmes
 * 
 */
public abstract class AudioDevice {
	protected double mTFSizeSecs;
	protected int mTFSizeSpls;
	private byte[] mRawBuff;
	private AudioFormat mAf;
	private long mNormOffset;
	private float mNormFactor;
	protected int mSampleSizeInBytes;

	public AudioDevice(double tfSize) {
		mTFSizeSecs = tfSize;
	}

	/**
	 * Has to be implemented by derived classes to provide the correct audio
	 * format.
	 * 
	 * @return The <tt>AudioFormat</tt> of the device.
	 */
	public abstract AudioFormat getFormat();

	/**
	 * Has to be called before any read operation on a device can take place.
	 * 
	 * @throws IOException
	 */
	public final void init() throws IOException {
		mAf = getFormat();
		if (mAf.getChannels() != 2) {
			throw new IOException(
					"AudioDevice.init(): Only stereo (2 channels) audio data accepted!");
		}
		int sFreq = Math.round(mAf.getSampleRate());
		mTFSizeSpls = (int) Math.round(mTFSizeSecs * sFreq);

		mRawBuff = new byte[mTFSizeSpls * mAf.getFrameSize()];

		mNormOffset = 1 << (mAf.getSampleSizeInBits() - 1);
		mNormFactor = 2.0f / ((1 << mAf.getSampleSizeInBits()) - 1.0f);
		mSampleSizeInBytes = mAf.getSampleSizeInBits() >> 3;
	}

	/**
	 * Read stereo audio data into the provided buffers. The <tt>float</tt> data
	 * will be normalized to the -1.0 ... .10 range according to the sample
	 * size. E.g. if we're recording 16 bit samples (amplitude range: -32768 ...
	 * 32767), -1.0 will correspond to -32768 and 1.0 will correspond to 32767.
	 * 
	 * @param leftBuff
	 *            The samples from the left channel will be stored in this
	 *            array. The length will have to be the constructor-specified
	 *            <tt>tfSize</tt>, converted to samples.
	 * @param rightBuff
	 *            The samples from the right channel will be stored in this
	 *            array. The length will have to be the constructor-specified
	 *            <tt>tfSize</tt>, converted to samples.
	 * @throws IOException
	 */
	public final void read(float[] leftBuff, float[] rightBuff)
			throws IOException {
		assert leftBuff.length == mTFSizeSpls : "Invalid sample buffer length";
		assert leftBuff.length == rightBuff.length : "Left sample buffer length != right sample buffer length";
		readRaw(mRawBuff);
		int bufIdx = 0;
		// assemble original samples from the individual bytes, normalize to
		// -1.0 ... 1.0, filter and apply FFT window
		for (int i = 0; i < mTFSizeSpls; ++i) {
			long tmpLeft = 0, tmpRight = 0;
			for (int j = mSampleSizeInBytes; j > 0; --j) {
				tmpLeft <<= 8;
				tmpLeft += mRawBuff[bufIdx + j - 1];
			}
			bufIdx += mSampleSizeInBytes;
			for (int j = mSampleSizeInBytes; j > 0; --j) {
				tmpRight <<= 8;
				tmpRight += mRawBuff[bufIdx + j - 1];
			}
			bufIdx += mSampleSizeInBytes;
			tmpLeft += mNormOffset;
			leftBuff[i] = tmpLeft * mNormFactor - 1.0f;
			tmpRight += mNormOffset;
			rightBuff[i] = tmpRight * mNormFactor - 1.0f;
		}
	}

	/**
	 * Derived classes will have to implement this method to read
	 * <tt>rawData.length</tt> raw samples from their actual device into the
	 * <tt>rawData</tt> <tt>byte</tt>-array.
	 * 
	 * @param rawData
	 * @throws IOException
	 */
	protected abstract void readRaw(byte[] rawData) throws IOException;

}
