/* SoundLocalizer - a sound source localization program
 * Copyright (C) 2010 Laurent Calmes
 *
 * This file is part of SoundLocalizer.
 *
 * SoundLocalizer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SoundLocalizer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SoundLocalizer.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.laurentcalmes.soundloc;

import java.util.Arrays;

/**
 * Audio filter base class.
 * 
 * @author Laurent Calmes
 * 
 */
public abstract class AudioFilter {

	public abstract float filterSample(float in);

	public abstract void reset();

	/**
	 * Computes the impulse response of the filter.
	 * 
	 * @return An array containing the impulse response (2048 samples).
	 */
	public final float[] impulseResponse() {
		float[] result = new float[2048];
		result[result.length >> 1] = 1.0f;
		for (int i = 0; i < result.length; ++i) {
			result[i] = filterSample(result[i]);
		}
		reset();
		return result;
	}

	/**
	 * Computes the frequency response of the filter. The FFT of the impulse
	 * response is computed and the lower half of the resulting complex array is
	 * returned (consecutive <tt>float</tt>s are the real and imaginary parts of
	 * the FFT points, respectively).
	 * 
	 * @return The lower half of the FFT of the impulse response (IR). As the IR
	 *         is real, the data in the upper half is redundant.
	 */
	public final float[] frequencyResponse() {
		float[] ir = impulseResponse();
		FFT fft = new FFT(ir.length);
		float[] fr = fft.compute(ir);
		return Arrays.copyOf(fr, fr.length >> 1);
	}

	/**
	 * Returns the normalized frequency (between 0.0 and 0.5) at which the
	 * cutoff of a lowpass filter - or the high cutoff of a bandpass filter -
	 * passes below -20 dB.
	 * 
	 * @return normalized cutoff frequency
	 */
	public final float maxNormFreq() {
		final float[] fr = frequencyResponse();
		int i = fr.length - 2;
		while ((i >= 0)
				&& (Math.sqrt(fr[i] * fr[i] + fr[i + 1] * fr[i + 1]) < 0.1)) {
			i -= 2;
		}
		i >>= 1;
		return (i < 0 ? 0.0f : (i + 1) / (float) 2048);
	}

	/**
	 * Returns the normalized frequency (between 0.0 and 0.5) at which the
	 * cutoff of a highpass filter - or the low cutoff of a bandpass filter -
	 * passes below -20 dB.
	 * 
	 * @return normalized cutoff frequency
	 */
	public final float minNormFreq() {
		final float[] fr = frequencyResponse();
		int i = 0;
		while ((i < fr.length)
				&& (Math.sqrt(fr[i] * fr[i] + fr[i + 1] * fr[i + 1]) < 0.1)) {
			i += 2;
		}
		i >>= 1;
		return (i > 1024 ? 0.5f : i / (float) 2048);
	}

}
