/* SoundLocalizer - a sound source localization program
 * Copyright (C) 2010 Laurent Calmes
 *
 * This file is part of SoundLocalizer.
 *
 * SoundLocalizer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SoundLocalizer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SoundLocalizer.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.laurentcalmes.soundloc;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JPanel;

/**
 * This class implements drawing of sound source azimuths.
 * 
 * @author Laurent Calmes
 * 
 */
public class AzDisplay extends JPanel implements ComponentListener {
	private static final long serialVersionUID = -7408736470890791891L;
	private BufferedImage mBgImg;
	private int mWinW = 800, mWinH = 600;
	private int mCircD = 600;
	private float mFontSize = 20.0f;
	private int mPtSize1 = 5, mPtSize2 = 3, mPtSize3 = 2;
	private int mMajTickLen = 20, mMinTickLen = 10;
	private int mYOffset = 10;
	private int mAzLblDist;
	private Font mFont;
	private List<LocalizationAlgorithm.Az> mPeakList = new ArrayList<LocalizationAlgorithm.Az>(
			0);

	public AzDisplay() {
		addComponentListener(this);
		setBackground(Color.black);
		setOpaque(true);
		setPreferredSize(new Dimension(800, 600));
	}

	/**
	 * Provide new sound source azimuth data to be drawn.
	 * 
	 * @param azs
	 *            Contains the new data to be drawn.
	 */
	public void updatePeaks(List<LocalizationAlgorithm.Az> azs) {
		mPeakList = azs;
		repaint();
	}

	/**
	 * Clear all azimuths on the screen.
	 */
	public void clear() {
		mPeakList.clear();
		repaint();
	}

	@Override
	public void paintComponent(Graphics g) {
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_RENDERING,
				RenderingHints.VALUE_RENDER_SPEED);
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
		g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
				RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
		g2.drawImage(mBgImg, 0, 0, null);
		drawAzs(g2, mPeakList);
	}

	@Override
	public void componentHidden(ComponentEvent e) {
	}

	@Override
	public void componentMoved(ComponentEvent e) {
	}

	@Override
	public void componentResized(ComponentEvent e) {
		mWinW = this.getWidth();
		mWinH = this.getHeight();
		if (mWinW >= (2 * mWinH)) {
			mCircD = Math.round(0.7f * mWinH);
			mFontSize = mWinH / 30.0f;
		} else {
			mCircD = Math.round(0.7f * mWinW);
			mFontSize = mWinW / 30.0f;
		}
		mPtSize1 = Math.round(mCircD / 120.0f);
		mPtSize1 = Math.max(1, mPtSize1);
		mPtSize2 = Math.round(mCircD / 200.0f);
		mPtSize2 = Math.max(1, mPtSize2);
		mPtSize3 = Math.round(mCircD / 300.0f);
		mPtSize3 = Math.max(1, mPtSize3);
		mMajTickLen = Math.round(mCircD / 30.0f);
		mMajTickLen = Math.max(1, mMajTickLen);
		mMinTickLen = Math.round(mCircD / 60.0f);
		mMinTickLen = Math.max(1, mMinTickLen);
		mFontSize = Math.max(mFontSize, 9.0f);
		mBgImg = new BufferedImage(mWinW, mWinH, BufferedImage.TYPE_4BYTE_ABGR);
		Graphics2D g2 = mBgImg.createGraphics();
		g2.setRenderingHint(RenderingHints.KEY_RENDERING,
				RenderingHints.VALUE_RENDER_QUALITY);
		g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
				RenderingHints.VALUE_INTERPOLATION_BICUBIC);
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
		g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
				RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
		mFont = g2.getFont().deriveFont(mFontSize);
		g2.setFont(mFont);
		FontMetrics fm = g2.getFontMetrics();
		Rectangle2D txtBox = fm.getStringBounds("-90", g2);
		mAzLblDist = Math.round((float) txtBox.getWidth() * 1.1f);
		mYOffset = Math.round((float) (txtBox.getHeight() - fm.getDescent()));
		drawBGimg(g2);
		g2.dispose();
		repaint();
	}

	@Override
	public void componentShown(ComponentEvent e) {
	}

	private void drawAzs(Graphics2D g2, List<LocalizationAlgorithm.Az> azs) {
		g2.setFont(mFont);
		for (LocalizationAlgorithm.Az az : azs) {
			AffineTransform saveTransfrm = g2.getTransform();
			g2.setPaint(Color.red);
			g2.setStroke(new BasicStroke(mPtSize2));
			g2.translate(mWinW / 2, mYOffset);
			double deg2rad = Math.PI / 180.0;
			double rotAngl = (90.0 + az.getAzimuth()) * deg2rad;
			g2.rotate(rotAngl);
			g2.drawLine(0, 0, (int) (mCircD * 0.5f * az.getHeight()), 0);
			g2.translate(mCircD / 2 + mAzLblDist, 0);
			g2.setStroke(new BasicStroke(mPtSize3));
			g2.setPaint(Color.white);
			g2.drawLine(0, 0, -mAzLblDist, 0);
			g2.rotate(-rotAngl);
			String azText = String.format("%1.1f", az.getAzimuth()) + "";
			FontMetrics fm = g2.getFontMetrics();
			Rectangle2D txtBox = fm.getStringBounds(azText, g2);
			txtBox.setRect(txtBox.getX(), txtBox.getY(), txtBox.getWidth() + 8,
					txtBox.getHeight() + 8);
			g2.translate(0, fm.getAscent());
			g2.draw(txtBox);
			g2.setPaint(Color.black);
			g2.fill(txtBox);
			g2.setPaint(Color.white);
			g2.drawString(azText, 4, 4);
			String hText = String.format("%.2f", az.getHeight() * 100) + "%";
			g2.translate(0, txtBox.getHeight() + 2);
			g2.setPaint(Color.red);
			g2.drawString(hText, 0, 0);
			g2.setTransform(saveTransfrm);
		}
	}

	private void drawBGimg(Graphics2D g2) {
		g2.setStroke(new BasicStroke(mPtSize1));
		g2.setPaint(Color.black);
		g2.fillRect(0, 0, mWinW, mWinH);
		g2.setPaint(Color.white);
		g2.setFont(mFont);
		g2.drawOval((mWinW - mCircD) / 2, -mCircD / 2 + mYOffset, mCircD,
				mCircD);
		g2.translate(mWinW / 2, mYOffset);
		AffineTransform saveTrnsfrm = g2.getTransform();
		double deg2rad = Math.PI / 180.0;
		for (int i = 180; i > 90; i -= 10) {
			g2.drawLine(mCircD / 2, 0, mCircD / 2 - mMajTickLen, 0);
			String label = String.valueOf(90 - i) + "";
			Rectangle2D txtBox = g2.getFontMetrics().getStringBounds(label, g2);
			g2.drawString(label,
					mCircD / 2 + Math.round(txtBox.getWidth() / 7), 5);
			g2.rotate(10 * deg2rad);
		}
		g2.setTransform(saveTrnsfrm);
		for (int i = 0; i < 90; i += 10) {
			g2.drawLine(-mCircD / 2, 0, -mCircD / 2 + mMajTickLen, 0);
			String label = String.valueOf(90 - i) + "";
			Rectangle2D txtBox = g2.getFontMetrics().getStringBounds(label, g2);
			double txtW = txtBox.getWidth();
			g2.drawString(label, -mCircD / 2 - Math.round(txtW + txtW / 7), 6);
			g2.rotate(-10 * deg2rad);
		}
		g2.setTransform(saveTrnsfrm);
		g2.drawLine(0, mCircD / 2, 0, mCircD / 2 - mYOffset);
		String label = String.valueOf(0) + "";
		Rectangle2D txtBox = g2.getFontMetrics().getStringBounds(label, g2);
		double txtH = txtBox.getHeight();
		g2.drawString(label, -Math.round(txtBox.getWidth() / 4), mCircD / 2
				+ Math.round(txtH + txtH / 10));
		g2.rotate(5 * deg2rad);
		for (int i = -85; i <= 85; i += 10) {
			g2.drawLine(mCircD / 2, 0, mCircD / 2 - mMinTickLen, 0);
			g2.rotate(10 * deg2rad);
		}
		g2.setTransform(saveTrnsfrm);
		g2.drawLine(0, -mYOffset, 0, 0);
	}
}
