/* SoundLocalizer - a sound source localization program
 * Copyright (C) 2010 Laurent Calmes
 *
 * This file is part of SoundLocalizer.
 *
 * SoundLocalizer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SoundLocalizer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SoundLocalizer.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.laurentcalmes.soundloc;

/**
 * Implements interaural time differences-based binaural sound source
 * localization using the correlation in the time domain algorithm.
 * 
 * @author Laurent Calmes
 * 
 */
public class CorrelationAlgorithm extends LocalizationAlgorithm {
	private float mIntegrationT; // integration time constant in samples
	private float[] mCorr;

	/**
	 * @param baseline
	 *            microphone distance in metres
	 * @param integrationT
	 *            integration time constant T in seconds
	 * @param tfSize
	 *            timeframe size in seconds
	 * @param ad
	 *            Audio device
	 */
	public CorrelationAlgorithm(double baseline, double integrationT,
			double tfSize, AudioDevice ad) {
		super(baseline, tfSize, ad);
		mIntegrationT = (float) integrationT;
	}

	@Override
	protected void initAlg() {
		mIntegrationT = (float) Math.exp(-1.0
				/ (mIntegrationT * mADevice.getFormat().getSampleRate()));
		mCorr = new float[2 * mMaxITDspls + 1];
	}

	@Override
	protected float[] initAzVals() {
		float[] ret = new float[2 * mMaxITDspls + 1];
		float rad2Deg = (float) (180.0 / Math.PI);
		for (int tau = -mMaxITDspls; tau <= mMaxITDspls; ++tau) {
			ret[tau + mMaxITDspls] = (float) Math.asin((soundSpeed * tau)
					/ (mADevice.getFormat().getSampleRate() * mBaseline));
			ret[tau + mMaxITDspls] *= rad2Deg;
		}
		return ret;
	}

	@Override
	protected float[] execute(float[] left, float[] right) {
		assert left.length == right.length : "Left and right channels must have the same length!";
		assert left.length - mMaxITDspls > mMaxITDspls : "Signals too short!";
		assert left.length == mTFSizeSpls : "Input signals must have length equal to timeframe size";

		for (int n = mMaxITDspls; n < mTFSizeSpls - mMaxITDspls; ++n) {
			for (int tau = -mMaxITDspls; tau <= mMaxITDspls; ++tau) {
				mCorr[tau + mMaxITDspls] *= mIntegrationT;
				mCorr[tau + mMaxITDspls] += left[n] * right[n - tau];
			}
		}

		float min = Float.MAX_VALUE, max = -Float.MAX_VALUE;

		for (float x : mCorr) {
			if (x >= max) {
				max = x;
			} else if (x < min) {
				min = x;
			}
		}

		// normalize to 0 ... 1
		float normfac = 1.0f / (max - min);
		for (int i = 0; i < mCorr.length; ++i)
			mCorr[i] = (mCorr[i] - min) * normfac;

		return mCorr;
	}

}
