/* SoundLocalizer - a sound source localization program
 * Copyright (C) 2010 Laurent Calmes
 *
 * This file is part of SoundLocalizer.
 *
 * SoundLocalizer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SoundLocalizer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SoundLocalizer.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.laurentcalmes.soundloc;

import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.Arrays;

/**
 * Implements interaural time differences-based binaural sound source
 * localization using the dual delay-line algorithm.
 * 
 * @author Laurent Calmes
 * 
 */
public class DDLAlgorithm extends LocalizationAlgorithm {
	private int mNDelays; // number of delays in each delay-line
	private int mFFTSize;
	private FFT mFFt;
	private int mFFTSize_2; // mFFTSize/2
	private float[] mDelayFactors; // table of precomputed complex phase delays
	private float[] mCoincMap; // 3D coincidence map
	private float mBeta; // coincidence time integration constant
	private float[] mWindow; // FFT window
	private float[] mLeftIn, mRightIn; // buffers for windowed left and right

	// mic

	// signals

	/**
	 * @param baseline
	 *            microphone distance in metres
	 * @param nDelays
	 *            number of delays in the delay line
	 * @param beta
	 *            coincidence map time integration time constant in seconds
	 * @param fftSize
	 *            size of the FFT in samples
	 * @param tfSize
	 *            timeframe size in seconds
	 * @param ad
	 *            Audio device
	 */
	public DDLAlgorithm(double baseline, int nDelays, double beta, int fftSize,
			double tfSize, AudioDevice ad) {
		super(baseline, tfSize, ad);
		mNDelays = nDelays;
		mFFTSize = fftSize;
		mBeta = (float) beta;
	}

	@Override
	protected void initAlg() throws FileNotFoundException {
		if (mFFTSize < mTFSizeSpls) {
			throw new IllegalArgumentException(
					"DDLAlgorithm.initAlg(): Size of FFT smaller than size of timeframe!");
		}

		mLeftIn = new float[mTFSizeSpls];
		mRightIn = new float[mTFSizeSpls];

		// init Hamming window
		mWindow = new float[mTFSizeSpls];
		final double factor = 2.0 * Math.PI / (mTFSizeSpls - 1);
		double f = 0.0;
		for (int i = 0; i < mTFSizeSpls; ++i) { // compute Hamming window
			mWindow[i] = 0.54f - 0.46f * (float) Math.cos(f);
			f += factor;
		}

		mFFt = new FFT(mFFTSize);
		mFFTSize_2 = mFFTSize >> 1;

		// init complex phase delays
		mDelayFactors = new float[2 * (mNDelays * mFFTSize_2)];
		mCoincMap = new float[mNDelays * mFFTSize_2];
		for (int i = 0; i < mNDelays; ++i) {
			// delay value in seconds
			double delay = mMaxITDsecs / 2.0
					* Math.sin(i / (mNDelays - 1.0) * Math.PI - Math.PI / 2.0);
			for (int j = 0; j < mFFTSize_2; ++j) {
				// frequency (in Hz) corresponding to FFT index j
				double frequency = j * mADevice.getFormat().getSampleRate()
						/ ((double) mFFTSize);
				// phase angle corresponding to delay value
				double phaseAngle = -2.0 * Math.PI * frequency * delay;
				// store real part
				mDelayFactors[(i * mFFTSize_2 + j) << 1] = (float) Math
						.cos(phaseAngle);
				// store imaginary part
				mDelayFactors[((i * mFFTSize_2 + j) << 1) + 1] = (float) Math
						.sin(phaseAngle);
			}
		}

		if (mLogging) {

			PrintStream file = new PrintStream(mLogPath + "DDL_DelayPhases.log");

			file.println("#");
			file
			.println("# Column 1: azimuth (degrees); Column2: frequency (Hz); Column 3: unwrapped phase");
			file.println("#");
			
			
			final double period = 2.0 * Math.PI;
			final double invPeriod = 1.0 / period;
			final double freqFac = mADevice.getFormat().getSampleRate()
					/ ((double) mFFTSize);

			for (int i = 0; i < mNDelays; ++i) {
				double oldPhase = 0.0;
				for (int j = 0; j < mFFTSize_2; ++j) {
					double freq = j * freqFac;
					file.print(String.valueOf(mAzVals[i]) + '\t');
					file.print(String.valueOf(freq) + '\t');
					float re = mDelayFactors[(i * mFFTSize_2 + j) << 1];
					float im = mDelayFactors[((i * mFFTSize_2 + j) << 1) + 1];
					double phase = Math.atan2(im, re);
					long nPeriod = Math.round((oldPhase - phase) * invPeriod);
					oldPhase = phase + nPeriod * period;
					file.println(String.valueOf(oldPhase));
				}
				file.println();
			}

		}
	}

	@Override
	protected float[] initAzVals() {
		float[] ret = new float[mNDelays];
		for (int i = 0; i < mNDelays; ++i) {
			ret[mNDelays - 1 - i] = 90.0f - i / (mNDelays - 1.0f) * 180.0f;
		}
		return ret;
	}

	@Override
	protected float[] execute(float[] left, float[] right)
			throws FileNotFoundException {
		assert left.length == right.length : "Left and right channels must have the same length!";
		assert left.length == mTFSizeSpls : "Input signals must have length equal to timeframe size";
		// apply window to input samples
		for (int i = 0; i < mTFSizeSpls; ++i) {
			mLeftIn[i] = left[i] * mWindow[i];
			mRightIn[i] = right[i] * mWindow[i];
		}

		// extend length of input to FFT size (padding with 0)
		float[] leftFFT = mFFt.compute(Arrays.copyOf(mLeftIn, mFFTSize));
		float[] rightFFT = mFFt.compute(Arrays.copyOf(mRightIn, mFFTSize));

		for (int i = 0; i < mNDelays; ++i) {
			for (int j = 0; j < mFFTSize_2; ++j) {
				final float ReL = leftFFT[j << 1], ImL = leftFFT[(j << 1) + 1];
				final float ReR = rightFFT[j << 1], ImR = rightFFT[(j << 1) + 1];
				final float ReD = mDelayFactors[(i * mFFTSize_2 + j) << 1];
				final float ImD = mDelayFactors[((i * mFFTSize_2 + j) << 1) + 1];
				final float delRe = (ReL - ReR) * ReD - (ImL + ImR) * ImD;
				final float delIm = (ImL - ImR) * ReD + (ReL + ReR) * ImD;
				mCoincMap[i * mFFTSize_2 + j] = mCoincMap[i * mFFTSize_2 + j]
						* mBeta
						+ (float) Math.sqrt(delRe * delRe + delIm * delIm);
			}
		}

		if (mLogging) {
			PrintStream file = new PrintStream(mLogPath + "DDL_CoincMap_"
					+ mTFNum + ".log");
			final double freqFac = mADevice.getFormat().getSampleRate()
					/ ((double) mFFTSize);
			file.println("#");
			file
			.println("# Column 1: azimuth (degrees); Column2: frequency (Hz); Column 3: dissimilarity");
			file.println("#");
			for (int i = 0; i < mNDelays; ++i) {
				for (int j = 0; j < mFFTSize_2; ++j) {
					file.println(String.valueOf(mAzVals[i]) + '\t' + j
							* freqFac + '\t' + mCoincMap[i * mFFTSize_2 + j]);
				}
				file.println();
			}
		}

		float[] result = new float[mNDelays];

		float min = Float.MAX_VALUE, max = -Float.MAX_VALUE;

		// integrate coincidence map over frequency & determine min. and max.
		for (int i = 0; i < mNDelays; ++i) {
			for (int j = 0; j < mFFTSize_2; ++j) {
				result[mNDelays - 1 - i] += mCoincMap[i * mFFTSize_2 + j];
			}
			if (result[mNDelays - 1 - i] >= max) {
				max = result[mNDelays - 1 - i];
			} else if (result[mNDelays - 1 - i] < min) {
				min = result[mNDelays - 1 - i];
			}
		}

		// invert and normalize to 0 ... 1
		float normfac = 1.0f / (max - min);
		for (int i = 0; i < result.length; ++i)
			result[i] = (max - result[i]) * normfac;

		return result;
	}
}
