/* SoundLocalizer - a sound source localization program
 * Copyright (C) 2010 Laurent Calmes
 *
 * This file is part of SoundLocalizer.
 *
 * SoundLocalizer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SoundLocalizer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SoundLocalizer.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.laurentcalmes.soundloc;

/**
 * A class for performing FFTs.
 * 
 * Implements the classic radix-2 FFT algorithm.
 * 
 * @author Laurent Calmes
 * 
 */
public class FFT {

	/**
	 * Integer-compute the base 2 logarithm of <tt>x</tt>.
	 * 
	 * @param x
	 *            the number from which to compute the base 2 logarithm
	 * @return the base 2 logarithm of <tt>x</tt>
	 */
	public static int log2(int x) {
		if (x == 0)
			throw new IllegalArgumentException("FFT.log2(): Argument == 0!");
		--x; // get the max index, x-1
		int mask = 1;
		int i = 0;
		while (x != 0) {
			x = x & (~mask); // AND off a bit
			mask <<= 1;
			++i;
		}
		return i; // return Log2 if all zero
	}

	private int mSize;
	private int mLogSize;
	private float[] mLookUpTable;

	/**
	 * Constructs an FFT object.
	 * 
	 * The FFTs performed by this FFT object will have size <tt>size</tt>.
	 * 
	 * @param size
	 *            the FFT size (must be a power of 2)
	 */
	public FFT(int size) {
		mLogSize = log2(size);
		if ((1 << mLogSize) != size) {
			throw new IllegalArgumentException(
					"FFT.FFT(): FFT size not power of 2!");
		}
		mSize = size;
		int le = size >> 1;
		mLookUpTable = new float[2 * (le - 1)];
		double theta = Math.PI / le;
		double[] wrecur = { Math.cos(theta), -Math.sin(theta) };
		double[] w = { wrecur[0], wrecur[1] };
		for (int i = 0; i < (le - 1); ++i) {
			mLookUpTable[2 * i] = (float) wrecur[0];
			mLookUpTable[2 * i + 1] = (float) wrecur[1];
			wrecur[0] = wrecur[0] * w[0] - wrecur[1] * w[1];
			wrecur[1] = wrecur[1] * w[0] + wrecur[0] * w[1];
		}
	}

	/**
	 * Returns the FFT size.
	 * 
	 * @return FFT size (is a power of 2)
	 */
	public int getSize() {
		return mSize;
	}

	/**
	 * Compute the FFT of the input <tt>data</tt>.
	 * 
	 * @param data
	 *            a real-valued <tt>float</tt> array with
	 *            <tt>data.length ==</tt> FFT size
	 * @return a complex <tt>float</tt> array containing the FFT of
	 *         <tt>data</tt>
	 */
	public float[] compute(float[] data) {
		assert data.length == mSize : "Data length != FFT length!";
		float[] result = new float[mSize << 1];
		for (int i = 0; i < data.length; ++i) {
			result[i << 1] = data[i];
			result[(i << 1) + 1] = 0.0f;
		}
		for (int l = 0, le = mSize, windex = 1; l < mLogSize; ++l, windex <<= 1) {
			le >>= 1;
			// first iteration with no multiplies
			for (int i = 0; i < mSize; i += (le << 1)) {
				float[] temp = { result[i << 1] + result[(i + le) << 1],
						result[(i << 1) + 1] + result[((i + le) << 1) + 1] };
				result[(i + le) << 1] = result[i << 1] - result[(i + le) << 1];
				result[((i + le) << 1) + 1] = result[(i << 1) + 1]
						- result[((i + le) << 1) + 1];
				result[i << 1] = temp[0];
				result[(i << 1) + 1] = temp[1];
			}
			// remaining iterations use stored mLookUpTable
			int uindex = windex - 1;
			for (int j = 1; j < le; ++j, uindex += windex) {
				float[] u = { mLookUpTable[uindex << 1],
						mLookUpTable[(uindex << 1) + 1] };
				for (int i = j; i < mSize; i += (le << 1)) {
					float[] temp = { result[i << 1] + result[(i + le) << 1],
							result[(i << 1) + 1] + result[((i + le) << 1) + 1] };
					float[] tm = { result[i << 1] - result[(i + le) << 1],
							result[(i << 1) + 1] - result[((i + le) << 1) + 1] };
					result[(i + le) << 1] = tm[0] * u[0] - tm[1] * u[1];
					result[((i + le) << 1) + 1] = tm[1] * u[0] + tm[0] * u[1];
					result[i << 1] = temp[0];
					result[(i << 1) + 1] = temp[1];
				}
			}
		}
		// rearrange data by bit reversing
		for (int i = 1, j = 0; i < (mSize - 1); ++i) {
			int k = mSize >> 1;
			while (k <= j) {
				j -= k;
				k >>= 1;
			}
			j += k;
			if (i < j) { // swap result[i] and result[j]
				float[] temp = { result[i << 1], result[(i << 1) + 1] };
				result[i << 1] = result[j << 1];
				result[(i << 1) + 1] = result[(j << 1) + 1];
				result[j << 1] = temp[0];
				result[(j << 1) + 1] = temp[1];
			}
		}
		return result;
	}
}
