/* SoundLocalizer - a sound source localization program
 * Copyright (C) 2010 Laurent Calmes
 *
 * This file is part of SoundLocalizer.
 *
 * SoundLocalizer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SoundLocalizer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SoundLocalizer.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.laurentcalmes.soundloc;

/**
 * Implementation of an infinite impulse response (IIR) filter
 * 
 * @author Laurent Calmes
 * 
 */
public class IIRFilter extends AudioFilter {
	private float mGain;
	private BiQuadSection[] mSections;

	IIRFilter() {
		this(0.0f, new BiQuadSection[0]);
	}

	IIRFilter(float gain, BiQuadSection[] sections) {
		mGain = gain;
		mSections = new BiQuadSection[sections.length];
		for (int i = 0; i < sections.length; ++i) {
			mSections[i] = new BiQuadSection(sections[i]);
		}

	}

	IIRFilter(IIRFilter filt) {
		this(filt.mGain, filt.mSections);
	}

	@Override
	public float filterSample(float in) {
		float inT = in, lastOut = 0.0f;
		for (int i = 0; i < mSections.length; ++i) {
			final BiQuadSection bqs = mSections[i];
			final float temp = mGain * inT + bqs.mHistory[0] * bqs.mA1
					+ bqs.mHistory[1] * bqs.mA2;
			lastOut = temp;
			lastOut += bqs.mHistory[0] * bqs.mB1;
			lastOut += bqs.mHistory[1] * bqs.mB2;
			bqs.mHistory[1] = bqs.mHistory[0];
			bqs.mHistory[0] = temp;
			inT = lastOut;
		}
		return lastOut;
	}

	public void reset() {
		for (int i = 0; i < mSections.length; ++i) {
			BiQuadSection bqs = mSections[i];
			bqs.mHistory[0] = bqs.mHistory[1] = 0.0f;
		}
	}

	/**
	 * Implements a biquadratic section for a recursive filter.
	 * 
	 * @author Laurent Calmes
	 *
	 */
	public static class BiQuadSection {
		float mB1, mB2, mA1, mA2;
		float[] mHistory = new float[2];

		BiQuadSection() {
			this(0.0f, 0.0f, 0.0f, 0.0f);
		}

		BiQuadSection(float b1, float b2, float a1, float a2) {
			mB1 = b1;
			mB2 = b2;
			mA1 = -a1;
			mA2 = -a2;
			mHistory[0] = mHistory[1] = 0.0f;
		}

		BiQuadSection(BiQuadSection bqs) {
			mB1 = bqs.mB1;
			mB2 = bqs.mB2;
			mA1 = bqs.mA1;
			mA2 = bqs.mA2;
		}
	}

}
