/* SoundLocalizer - a sound source localization program
 * Copyright (C) 2010 Laurent Calmes
 *
 * This file is part of SoundLocalizer.
 *
 * SoundLocalizer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SoundLocalizer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SoundLocalizer.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.laurentcalmes.soundloc;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.sound.sampled.AudioFormat;

import com.laurentcalmes.soundloc.IIRFilter.BiQuadSection;

/**
 * Abstract sound source localization algorithms base class.
 * 
 * @author Laurent Calmes
 * 
 */
public abstract class LocalizationAlgorithm {
	public static final double soundSpeed = 340.0; // in m/s
	protected double mBaseline; // inter-microphone distance in m
	protected int mMaxITDspls; // maximal ITD in samples
	protected double mMaxITDsecs; // maximal ITD in seconds
	protected double mTFSizeSecs; // timeframe size in seconds
	protected int mTFSizeSpls; // timeframe size in samples
	protected float[] mAzVals; // index to azimuth mappings
	private float[] mLocFunc; // localization function

	protected boolean mLogging;
	protected String mLogPath;
	protected long mTFNum = 0;

	protected AudioDevice mADevice;
	protected float[] mLeftBuff, mRightBuff; // left & right mic. signal buffers
	private AudioFilter mFilterLeft, mFilterRight; // left and right channel

	// filters

	/**
	 * The Az class stores azimuth and associated confidence (peak height) of
	 * detected sound sources.
	 * 
	 * @author Laurent Calmes
	 * 
	 */
	public static class Az implements Comparable<Az> {
		private float mAzimuth, mHeight;

		Az() {
			this(0.0f, 0.0f);
		}

		Az(float azimuth, float height) {
			mAzimuth = azimuth;
			mHeight = height;
		}

		Az(Az a) {
			this(a.mAzimuth, a.mHeight);
		}

		/**
		 * Return the azimuth value.
		 * 
		 * @return the azimuth to a sound source
		 */
		public float getAzimuth() {
			return mAzimuth;
		}

		/**
		 * Return the height (algorithm-dependent confidence) value.
		 * 
		 * @return the height value.
		 */
		public float getHeight() {
			return mHeight;
		}

		@Override
		public boolean equals(Object obj) {
			if (obj == null)
				return false;

			if (obj == this)
				return true;

			if (!obj.getClass().equals(getClass()))
				return false;

			Az in = (Az) obj;

			return (this.mAzimuth == in.mAzimuth)
					&& (this.mHeight == in.mHeight);
		}

		@Override
		public String toString() {
			return new String("(" + this.mAzimuth + "; " + this.mHeight + ")");
		}

		@Override
		public int compareTo(Az o) {
			if ((this == o) || (this.equals(o)))
				return 0;
			else
				return (this.mHeight - o.mHeight) < 0.0f ? -1 : 1;
		}
	}

	/**
	 * @param baseline
	 *            microphone distance in metres
	 * @param tfSize
	 *            timeframe size in seconds
	 * @param ad
	 *            <tt>AudioDevice</tt> used by the
	 *            <tt>LocalizationAlgorithm</tt>
	 */
	public LocalizationAlgorithm(double baseline, double tfSize, AudioDevice ad) {
		mADevice = ad;
		mBaseline = baseline;
		mMaxITDsecs = baseline / soundSpeed;

		AudioFormat format = ad.getFormat();

		int sFreq = Math.round(format.getSampleRate());

		mMaxITDspls = (int) Math.floor(mMaxITDsecs * sFreq);

		mTFSizeSecs = tfSize;
		mTFSizeSpls = (int) Math.round(mTFSizeSecs * sFreq);

		mLeftBuff = new float[mTFSizeSpls];
		mRightBuff = new float[mTFSizeSpls];

		switch (sFreq) {
		case 11025:
			mFilterLeft = new IIRFilter(filt11025);
			mFilterRight = new IIRFilter(filt11025);
			break;
		case 16000:
			mFilterLeft = new IIRFilter(filt16000);
			mFilterRight = new IIRFilter(filt16000);
			break;
		case 22050:
			mFilterLeft = new IIRFilter(filt22050);
			mFilterRight = new IIRFilter(filt22050);
			break;
		case 44100:
			mFilterLeft = new IIRFilter(filt44100);
			mFilterRight = new IIRFilter(filt44100);
			break;
		case 48000:
			mFilterLeft = new IIRFilter(filt48000);
			mFilterRight = new IIRFilter(filt48000);
			break;
		default:
			mFilterLeft = new DummyFilter();
			mFilterRight = new DummyFilter();
			break;
		}
	}

	/**
	 * Initialize the algorithm. Has to be called before calling
	 * <tt>compute()</tt>.
	 * 
	 */
	public final void init() throws FileNotFoundException {
		// right now, only need to initialize index to az mappings
		mAzVals = initAzVals();
		initAlg();
	}

	/**
	 * Derived classes can do initialization here.
	 * 
	 * @throws FileNotFoundException
	 */
	protected abstract void initAlg() throws FileNotFoundException;

	/**
	 * Reads audio data, calls the derived-class <tt>execute()</tt> and returns
	 * the detected sound sources.
	 * 
	 * @return the list of detected sound sources
	 * @throws IOException
	 */
	public final List<Az> compute() throws IOException {
		mADevice.read(mLeftBuff, mRightBuff);

		// bandpass filter audio
		for (int i = 0; i < mTFSizeSpls; ++i) {
			mLeftBuff[i] = mFilterLeft.filterSample(mLeftBuff[i]);
			mRightBuff[i] = mFilterRight.filterSample(mRightBuff[i]);
		}

		mLocFunc = execute(mLeftBuff, mRightBuff);

		if (mLogging) {
			PrintStream logFile = new PrintStream(mLogPath + "LocFunc_"
					+ String.valueOf(mTFNum) + ".log");
			logFile.println("#");
			logFile
					.println("# Column 1: azimuth (degrees); Column2: localization function value");
			logFile.println("#");
			for (int i = 0; i < mLocFunc.length; ++i) {
				logFile.println(String.valueOf(mAzVals[i]) + '\t'
						+ String.valueOf(mLocFunc[i]));
			}
			++mTFNum;
		}

		return pickPeaks(mLocFunc);
	}

	/**
	 * Enable / disable logging.
	 * 
	 * @param logging
	 *            <tt>true</tt>: logging enabled; <tt>false</tt>: logging
	 *            disabled
	 */
	public final void setLoggingEnabled(boolean logging) {
		mLogging = logging;
	}

	/**
	 * Return logging status.
	 * 
	 * @return <tt>true</tt>: logging enabled; <tt>false</tt>: logging disabled
	 */
	public final boolean getLoggingEnabled() {
		return mLogging;
	}

	/**
	 * Set path to log files.
	 * 
	 * @param path
	 *            Path to log files
	 */
	public final void setLogPath(String path) {
		String fileSep = System.getProperty("file.separator");
		mLogPath = path;
		if (!mLogPath.endsWith(fileSep)) {
			mLogPath = path + fileSep;
		}
	}

	/**
	 * Return path to log files.
	 * 
	 * @return Path to log files
	 */
	public final String getLogPath() {
		return mLogPath;
	}

	/**
	 * Returns the array of azimuth values detectable by the algorithm.
	 * 
	 * @see #initAzVals()
	 * 
	 * @return the array of possible azimuth values.
	 */
	protected final float[] getAzVals() {
		return mAzVals;
	}

	/**
	 * Returns the array of azimuth values detectable by the algorithm. The
	 * value at index 0 of the returned array is to the leftmost (closest to
	 * -90&deg;) detectable azimuth, while the last value (index
	 * <tt>length-1</tt>) is the rightmost azimuth value (closest to +90&deg;).
	 * 
	 * @return the array of possible azimuth values.
	 */
	protected abstract float[] initAzVals();

	/**
	 * The actual algorithm implementation. It should compute a localization
	 * function on the <tt>left</tt> and <tt>right</tt> input signals. Position
	 * in localization function array corresponds to azimuth (index 0 closest to
	 * -90&deg; and index <tt>length-1</tt> closest to +90&deg;). The higher a
	 * value at a given index, the higher the confidence for a sound source at
	 * that position. The values have to be normalized to the range 0.0 ... 1.0.
	 * 
	 * @param left
	 *            The left input signal.
	 * @param right
	 *            The right input signal.
	 * @return The localization function.
	 */
	protected abstract float[] execute(float[] left, float[] right)
			throws FileNotFoundException;

	private final List<Az> pickPeaks(float[] locFunc) {
		ArrayList<Az> peaks = new ArrayList<Az>(10);
		int i1 = locFunc.length - 1, i2 = 1;
		peaks.clear();
		for (int i = 0; i < locFunc.length; ++i) {
			if ((locFunc[i] > 0.5f) && (locFunc[i] > locFunc[i1])
					&& (locFunc[i] > locFunc[i2]))
				peaks.add(new Az(mAzVals[i], locFunc[i]));
			i1 = (i1 + 1) % locFunc.length;
			i2 = (i2 + 1) % locFunc.length;
		}
		Collections.sort(peaks, Collections.reverseOrder());
		if (peaks.size() > 10) {
			peaks.subList(10, peaks.size() - 1).clear();
		}
		return peaks;
	}

	// /////////////////////////////////////////////////////////////
	// the following band-pass filters are designed for a center //
	// frequency of about 2050Hz and a bandwidth of about 3.9kHz //
	// at various sampling rates //
	// /////////////////////////////////////////////////////////////

	/**
	 * 6 section (12th order) Butterworth IIR band pass filter; center
	 * frequency: 0.186; bandw: 0.354; for use at a sampling rate of 11.025kHz
	 */
	private final IIRFilter filt11025 = new IIRFilter(0.733893873446009f,
			new BiQuadSection[] {
					new BiQuadSection(0.0000000f, -1.0000000f, 0.7313905f,
							0.1475091f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.8910211f,
							0.8942025f),
					new BiQuadSection(0.0000000f, -1.0000000f, 0.8404634f,
							0.3042392f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9200414f,
							0.9231953f),
					new BiQuadSection(0.0000000f, -1.0000000f, 1.0922940f,
							0.6781853f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9686413f,
							0.9718035f) });

	/**
	 * 6 section (12th order) Butterworth IIR band pass filter; center
	 * frequency: 0.128125; bandw: 0.24375; for use at a sampling rate of 16kHz
	 */
	private final IIRFilter filt16000 = new IIRFilter(0.5451950303462342f,
			new BiQuadSection[] {
					new BiQuadSection(0.0000000f, -1.0000000f, -0.0385467f,
							0.0191662f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9228326f,
							0.9244329f),
					new BiQuadSection(0.0000000f, -1.0000000f, -0.0226965f,
							0.1817501f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9445046f,
							0.9460637f),
					new BiQuadSection(0.0000000f, -1.0000000f, -0.0039227f,
							0.6007223f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9790277f,
							0.9805571f) });

	/**
	 * 6 section (12th order) Butterworth IIR band pass filter; center
	 * frequency: 0.09297; bandw: 0.17687; for use at a sampling rate of
	 * 22.050kHz
	 */
	private final IIRFilter filt22050 = new IIRFilter(0.42310594406442875f,
			new BiQuadSection[] {
					new BiQuadSection(0.0000000f, -1.0000000f, -0.4871394f,
							0.0762034f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9434705f,
							0.9443306f),
					new BiQuadSection(0.0000000f, -1.0000000f, -0.5371357f,
							0.2333643f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9597541f,
							0.9605852f),
					new BiQuadSection(0.0000000f, -1.0000000f, -0.6856398f,
							0.6323152f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9851071f,
							0.9859176f) });

	/**
	 * 6 section (12th order) Butterworth IIR band pass filter; center
	 * frequency: 0.0465; bandw: 0.0884; for use at a sampling rate of 44.1kHz
	 */
	private final IIRFilter filt44100 = new IIRFilter(0.23769088446877812f,
			new BiQuadSection[] {
					new BiQuadSection(0.0000000f, -1.0000000f, -1.1416609f,
							0.3347037f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9710172f,
							0.9712439f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.2445259f,
							0.4662141f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9796050f,
							0.9798219f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.4878532f,
							0.7652840f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9926540f,
							0.9928633f) });

	/**
	 * 6 section (12th order) Butterworth IIR band pass filter; center
	 * frequency: 0.0427; bandw: 0.08125; for use at a sampling rate of 48kHz
	 */
	private final IIRFilter filt48000 = new IIRFilter(0.22088535127467304f,
			new BiQuadSection[] {
					new BiQuadSection(0.0000000f, -1.0000000f, -1.2005197f,
							0.3682837f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9738518f,
							0.9740362f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.3038064f,
							0.4954571f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9815998f,
							0.9817764f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.5433269f,
							0.7804966f),
					new BiQuadSection(0.0000000f, -1.0000000f, -1.9933848f,
							0.9935553f) });
}
