/* SoundLocalizer - a sound source localization program
 * Copyright (C) 2010 Laurent Calmes
 *
 * This file is part of SoundLocalizer.
 *
 * SoundLocalizer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SoundLocalizer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SoundLocalizer.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.laurentcalmes.soundloc;

import java.awt.Component;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.filechooser.FileNameExtensionFilter;

/**
 * SoundLocalizer UI.
 * 
 * @author Laurent Calmes
 *
 */
public class SoundLocalizerWindow extends JFrame {
	private static final long serialVersionUID = 3325918384617427035L;
	private JComboBox mSFreqCB, mHwCB, mSynCB, mSFreqSynCB;
	private ButtonGroup mAlgorithmBttnGrp;
	private JTextField mFileNameTF;
	private JSpinner mBaselineSpin, mSynAz, mSynSineFreq;
	private AzDisplay mAzDisp;
	private JCheckBox mEnableLoggingBttn;
	private JTextField mLogPathTF;

	private String mUseAudioDev;
	private boolean mIsFilenameSet = false, mIsLogPathSet = false;
	private Map<String, Mixer.Info> mAudioHWlst;
	private LocalizationAlgorithm mAlg = null;
	private LocalizerTask mLocTask;

	public SoundLocalizerWindow() {
		super("SoundLocalizer");

		setLayout(new GridBagLayout());

		GridBagConstraints gbTopCstrts = new GridBagConstraints();

		mAzDisp = new AzDisplay();

		gbTopCstrts.gridx = 0;
		gbTopCstrts.gridy = 0;
		gbTopCstrts.gridheight = 4;
		gbTopCstrts.weightx = 1;
		gbTopCstrts.weighty = 1;
		gbTopCstrts.anchor = GridBagConstraints.NORTH;
		gbTopCstrts.fill = GridBagConstraints.BOTH;

		add(mAzDisp, gbTopCstrts);

		AudioPanel audioPanel = new AudioPanel(this);

		gbTopCstrts.gridx = 1;
		gbTopCstrts.gridy = 0;
		gbTopCstrts.gridheight = 1;
		gbTopCstrts.weightx = 0;
		gbTopCstrts.weighty = 0;
		gbTopCstrts.fill = GridBagConstraints.HORIZONTAL;

		add(audioPanel, gbTopCstrts);

		GeneralPanel genPanel = new GeneralPanel(this);

		gbTopCstrts.gridx = 1;
		gbTopCstrts.gridy = 1;
		add(genPanel, gbTopCstrts);

		AlgorithmPanel algPanel = new AlgorithmPanel();

		gbTopCstrts.gridx = 1;
		gbTopCstrts.gridy = 2;

		add(algPanel, gbTopCstrts);

		ControlsPanel startStopCntrls = new ControlsPanel();

		gbTopCstrts.gridx = 0;
		gbTopCstrts.gridy = 4;

		add(startStopCntrls, gbTopCstrts);
	}

	private class AudioPanel extends JPanel {
		private static final long serialVersionUID = -2577235885772698568L;

		AudioPanel(Component parent) {
			super(new GridBagLayout());
			mFileNameTF = new JTextField("Enter filename here ...");
			mFileNameTF.setEnabled(false);
			mFileNameTF.setFont(mFileNameTF.getFont().deriveFont(Font.ITALIC));
			mFileNameTF.setToolTipText(mFileNameTF.getText()
					+ " or click button to the right");
			final JButton fileBtn = new JButton("...");
			fileBtn.setEnabled(false);
			fileBtn.setToolTipText("Click to select file");
			final Component prnt = parent;
			fileBtn.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					JFileChooser fc = new JFileChooser();
					FileNameExtensionFilter filter = new FileNameExtensionFilter(
							".wav audio files", "wav", "WAV");
					fc.setFileFilter(filter);
					if (fc.showOpenDialog(prnt) == JFileChooser.APPROVE_OPTION) {
						mIsFilenameSet = true;
						mFileNameTF.setFont(mFileNameTF.getFont().deriveFont(
								Font.PLAIN));
						mFileNameTF.setText(fc.getSelectedFile().getPath());
						mFileNameTF.setToolTipText(mFileNameTF.getText());
					}

				}
			});

			String[] sFreqItems = { "48000 Hz", "44100 Hz", "22050 Hz",
					"16000 Hz", "11025 Hz" };
			mSFreqCB = new JComboBox(sFreqItems);
			mSFreqCB.setEditable(false);
			mSFreqCB.setEnabled(false);
			mSFreqCB.setToolTipText("Sampling frequency");

			mAudioHWlst = enumerateAudioHW();
			mHwCB = new JComboBox(mAudioHWlst.keySet().toArray());
			mHwCB.setEditable(false);
			mHwCB.setEnabled(false);

			mSFreqSynCB = new JComboBox(sFreqItems);
			mSFreqSynCB.setEditable(false);
			mSFreqSynCB.setToolTipText("Sampling frequency");

			SynthDevice.SigType[] synItems = { SynthDevice.SigType.IMPULSE,
					SynthDevice.SigType.NOISE, SynthDevice.SigType.SINE };
			mSynCB = new JComboBox(synItems);
			mSynCB.setEditable(false);
			mSynCB.setToolTipText("Synthesized signal type");
			mSynCB.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					JComboBox cb = (JComboBox) e.getSource();
					if (cb.getSelectedItem().equals(SynthDevice.SigType.SINE)) {
						mSynSineFreq.setEnabled(true);
					} else {
						mSynSineFreq.setEnabled(false);
					}
				}
			});

			mSynAz = new JSpinner(new SpinnerNumberModel(31.3f, -90.0f, 90.0f,
					0.1f));
			mSynAz.setToolTipText("Virtual source azimuth");
			mSynAz.setEditor(new JSpinner.NumberEditor(mSynAz, "0.0°"));

			mSynSineFreq = new JSpinner(new SpinnerNumberModel(500.0f, 0.0f,
					24000.0f, 10.0f));
			mSynSineFreq
					.setToolTipText("Synthesized sine frequency (only for SINE signal type)");
			mSynSineFreq.setEditor(new JSpinner.NumberEditor(mSynSineFreq,
					"0 Hz"));
			if (mSynCB.getSelectedItem().equals(SynthDevice.SigType.SINE)) {
				mSynSineFreq.setEnabled(true);
			} else {
				mSynSineFreq.setEnabled(false);
			}

			JRadioButton audioHWBtn = new JRadioButton("Hardware device: ");
			audioHWBtn.setActionCommand("HW");
			JRadioButton audioWAVBtn = new JRadioButton("WAV file : ");
			mUseAudioDev = "SYN";
			audioWAVBtn.setActionCommand("WAV");
			JRadioButton audioSYNBtn = new JRadioButton("Synth device: ");
			audioSYNBtn.setActionCommand("SYN");
			audioSYNBtn.setSelected(true);
			ButtonGroup audioGroup = new ButtonGroup();
			audioGroup.add(audioHWBtn);
			audioGroup.add(audioWAVBtn);
			audioGroup.add(audioSYNBtn);

			ActionListener audioGrpListener = new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					if ("HW".equals(e.getActionCommand())) {
						mUseAudioDev = "HW";
						mSFreqCB.setEnabled(true);
						mHwCB.setEnabled(true);
						mFileNameTF.setEnabled(false);
						fileBtn.setEnabled(false);
						mSynCB.setEnabled(false);
						mSFreqSynCB.setEnabled(false);
						mSynAz.setEnabled(false);
						mSynSineFreq.setEnabled(false);
					} else if ("WAV".equals(e.getActionCommand())) {
						mUseAudioDev = "WAV";
						mSFreqCB.setEnabled(false);
						mHwCB.setEnabled(false);
						mFileNameTF.setEnabled(true);
						fileBtn.setEnabled(true);
						mSynCB.setEnabled(false);
						mSFreqSynCB.setEnabled(false);
						mSynAz.setEnabled(false);
						mSynSineFreq.setEnabled(false);
					} else if ("SYN".equals(e.getActionCommand())) {
						mUseAudioDev = "SYN";
						mSFreqCB.setEnabled(false);
						mHwCB.setEnabled(false);
						mFileNameTF.setEnabled(false);
						fileBtn.setEnabled(false);
						mSynCB.setEnabled(true);
						mSFreqSynCB.setEnabled(true);
						mSynAz.setEnabled(true);
						mSynSineFreq.setEnabled(true);
					}
				}
			};
			audioHWBtn.addActionListener(audioGrpListener);
			audioWAVBtn.addActionListener(audioGrpListener);
			audioSYNBtn.addActionListener(audioGrpListener);

			GridBagConstraints gbAudioCstrts = new GridBagConstraints();
			setBorder(BorderFactory.createTitledBorder(BorderFactory
					.createEtchedBorder(), "Audio setup"));

			gbAudioCstrts.gridx = 0;
			gbAudioCstrts.gridy = 0;
			gbAudioCstrts.weightx = 0.5;
			gbAudioCstrts.weighty = 0.5;
			gbAudioCstrts.anchor = GridBagConstraints.WEST;
			gbAudioCstrts.fill = GridBagConstraints.HORIZONTAL;
			add(audioHWBtn, gbAudioCstrts);

			gbAudioCstrts.anchor = GridBagConstraints.EAST;
			gbAudioCstrts.fill = GridBagConstraints.NONE;
			gbAudioCstrts.gridx = 1;
			add(mHwCB, gbAudioCstrts);

			gbAudioCstrts.gridx = 2;
			add(mSFreqCB, gbAudioCstrts);

			gbAudioCstrts.gridx = 0;
			gbAudioCstrts.gridy = 1;
			gbAudioCstrts.anchor = GridBagConstraints.WEST;
			gbAudioCstrts.fill = GridBagConstraints.HORIZONTAL;
			add(audioWAVBtn, gbAudioCstrts);

			gbAudioCstrts.anchor = GridBagConstraints.EAST;
			gbAudioCstrts.fill = GridBagConstraints.NONE;
			gbAudioCstrts.gridx = 1;
			add(mFileNameTF, gbAudioCstrts);

			gbAudioCstrts.gridx = 2;
			gbAudioCstrts.anchor = GridBagConstraints.WEST;
			add(fileBtn, gbAudioCstrts);

			gbAudioCstrts.gridx = 0;
			gbAudioCstrts.gridy = 2;
			gbAudioCstrts.anchor = GridBagConstraints.WEST;
			gbAudioCstrts.fill = GridBagConstraints.HORIZONTAL;
			add(audioSYNBtn, gbAudioCstrts);

			gbAudioCstrts.anchor = GridBagConstraints.EAST;
			gbAudioCstrts.fill = GridBagConstraints.NONE;
			gbAudioCstrts.gridx = 1;
			add(mSynCB, gbAudioCstrts);

			gbAudioCstrts.gridx = 2;
			add(mSFreqSynCB, gbAudioCstrts);

			gbAudioCstrts.gridx = 1;
			gbAudioCstrts.gridy = 3;
			gbAudioCstrts.anchor = GridBagConstraints.EAST;
			gbAudioCstrts.fill = GridBagConstraints.NONE;
			add(mSynSineFreq, gbAudioCstrts);

			gbAudioCstrts.gridx = 2;
			add(mSynAz, gbAudioCstrts);

		}

	}

	private Map<String, Mixer.Info> enumerateAudioHW() {
		Map<String, Mixer.Info> result = new HashMap<String, Mixer.Info>();
		for (Mixer.Info mixinf : AudioSystem.getMixerInfo()) {
			result.put(mixinf.getName(), mixinf);
		}
		return result;
	}

	private class GeneralPanel extends JPanel {
		private static final long serialVersionUID = -6899356721346088927L;

		GeneralPanel(Component parent) {
			super(new GridBagLayout());
			setBorder(BorderFactory.createTitledBorder(BorderFactory
					.createEtchedBorder(), "General setup"));
			GridBagConstraints gbGenCstrts = new GridBagConstraints();
			JLabel blLabel = new JLabel("Baseline: ", JLabel.LEFT);
			blLabel.setToolTipText("Distance between the 2 microphones");

			gbGenCstrts.gridx = 0;
			gbGenCstrts.gridy = 0;
			gbGenCstrts.weightx = 0.5;
			gbGenCstrts.weighty = 0.5;
			gbGenCstrts.anchor = GridBagConstraints.WEST;
			gbGenCstrts.fill = GridBagConstraints.HORIZONTAL;
			add(blLabel, gbGenCstrts);

			mBaselineSpin = new JSpinner(new SpinnerNumberModel(20.5f, 0.0f,
					100.0f, 0.1f));
			mBaselineSpin.setToolTipText("Distance between the 2 microphones");
			mBaselineSpin.setEditor(new JSpinner.NumberEditor(mBaselineSpin,
					"0.0 cm"));

			gbGenCstrts.gridx = 1;
			gbGenCstrts.gridy = 0;
			gbGenCstrts.anchor = GridBagConstraints.EAST;
			add(mBaselineSpin, gbGenCstrts);

			mLogPathTF = new JTextField("Enter path ...");
			mLogPathTF.setEnabled(false);
			mLogPathTF.setFont(mLogPathTF.getFont().deriveFont(Font.ITALIC));
			mLogPathTF.setToolTipText(mLogPathTF.getText()
					+ " or click button to the right");
			final JButton fileBtn = new JButton("...");
			fileBtn.setEnabled(false);
			fileBtn.setToolTipText("Click to select file");
			final Component prnt = parent;
			fileBtn.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					JFileChooser fc = new JFileChooser();
					fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
					if (fc.showOpenDialog(prnt) == JFileChooser.APPROVE_OPTION) {
						mIsLogPathSet = true;
						mLogPathTF.setFont(mLogPathTF.getFont().deriveFont(
								Font.PLAIN));
						mLogPathTF.setText(fc.getSelectedFile().getPath());
						mLogPathTF.setToolTipText(mLogPathTF.getText());
					}

				}
			});

			mEnableLoggingBttn = new JCheckBox("Enable logging", false);
			mEnableLoggingBttn.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					if (mEnableLoggingBttn.getSelectedObjects() == null) {
						mLogPathTF.setEnabled(false);
						fileBtn.setEnabled(false);
					} else {
						mLogPathTF.setEnabled(true);
						fileBtn.setEnabled(true);
					}
				}
			});

			gbGenCstrts.gridx = 0;
			gbGenCstrts.gridy = 1;
			gbGenCstrts.weightx = 0.5;
			gbGenCstrts.weighty = 0.5;
			gbGenCstrts.anchor = GridBagConstraints.WEST;
			gbGenCstrts.fill = GridBagConstraints.HORIZONTAL;
			add(mEnableLoggingBttn, gbGenCstrts);

			gbGenCstrts.gridx = 1;
			gbGenCstrts.gridy = 1;
			gbGenCstrts.anchor = GridBagConstraints.EAST;
			gbGenCstrts.fill = GridBagConstraints.NONE;
			add(mLogPathTF, gbGenCstrts);

			gbGenCstrts.gridx = 2;
			gbGenCstrts.gridy = 1;
			gbGenCstrts.anchor = GridBagConstraints.WEST;
			add(fileBtn, gbGenCstrts);
		}
	}

	private class AlgorithmPanel extends JPanel {
		private static final long serialVersionUID = 517609382773719324L;

		AlgorithmPanel() {
			super(new GridLayout(0, 1));
			setBorder(BorderFactory.createTitledBorder(BorderFactory
					.createEtchedBorder(), "Algorithms"));
			JRadioButton corrBtn = new JRadioButton("Cross-correlation");
			corrBtn.setActionCommand("corr");
			JRadioButton ddlBtn = new JRadioButton("Dual delay-line");
			ddlBtn.setActionCommand("ddl");
			ddlBtn.setSelected(true);
			mAlgorithmBttnGrp = new ButtonGroup();
			mAlgorithmBttnGrp.add(corrBtn);
			mAlgorithmBttnGrp.add(ddlBtn);
			add(corrBtn);
			add(ddlBtn);
		}
	}

	private void setupAlgorithm() throws IOException,
			UnsupportedAudioFileException, LineUnavailableException {

		AudioDevice ad = null;

		final double tfSize = 62e-3; // timeframe size (62 ms)

		if ("HW".equals(mUseAudioDev)) {
			String sFreqStr = (String) mSFreqCB.getSelectedItem();
			int sFreq = Integer.valueOf(sFreqStr.substring(0,
					sFreqStr.length() - 3));
			ad = new HWDevice(tfSize,
					new AudioFormat(sFreq, 16, 2, true, false), mAudioHWlst
							.get(mHwCB.getSelectedItem()));
		} else if ("WAV".equals(mUseAudioDev)) {
			if (!mIsFilenameSet) {
				throw new UnsupportedAudioFileException(
						"SoundLocalizerWindow.setupAlgorithm(): No WAV file has been set!");
			}
			ad = new WavDevice(tfSize, mFileNameTF.getText());
		} else if ("SYN".equals(mUseAudioDev)) {
			SynthDevice.SigType sigtype = (SynthDevice.SigType) mSynCB
					.getSelectedItem();
			String sFreqStr = (String) mSFreqSynCB.getSelectedItem();
			int sFreq = Integer.valueOf(sFreqStr.substring(0,
					sFreqStr.length() - 3));
			double az = (Double) mSynAz.getValue();
			double sineFreq = (Double) mSynSineFreq.getValue();
			double baseline = (Double) mBaselineSpin.getValue() / 100;
			SynthDevice sd = new SynthDevice(sigtype, az, baseline, tfSize,
					new AudioFormat(sFreq, 16, 2, true, false));
			sd.setSineFreq((float) sineFreq);
			ad = sd;
		}
		ad.init(); // don't forget to initialize the AudioDevice!

		// baseline, converted from cm to m
		double baseline = (Double) mBaselineSpin.getValue() / 100;

		String algorithm = mAlgorithmBttnGrp.getSelection().getActionCommand();

		if ("corr".equals(algorithm)) {
			mAlg = new CorrelationAlgorithm(baseline, 10e-3, tfSize, ad);

		} else if ("ddl".equals(algorithm)) {
			final int tfSizeSpls = (int) Math.round(tfSize
					* ad.getFormat().getSampleRate());
			final int logSize = FFT.log2(tfSizeSpls);
			int fftSize = 1 << logSize;
			if (fftSize < tfSizeSpls) {
				fftSize <<= 1;
			}
			mAlg = new DDLAlgorithm(baseline, 361, 0.88, fftSize, tfSize, ad);
		}
		if (mEnableLoggingBttn.getSelectedObjects() != null) {
			if (mIsLogPathSet) {
				mAlg.setLoggingEnabled(true);
				mAlg.setLogPath(mLogPathTF.getText());
			} else {
				throw new IOException(
						"SoundLocalizerWindow.setupAlgorithm(): No log file path has been set!");

			}
		}
		mAlg.init(); // don't forget to initialize the Algorithm!
	}

	private class ControlsPanel extends JPanel {
		private static final long serialVersionUID = -7936090066397590160L;

		ControlsPanel() {
			super(new GridBagLayout());
			GridBagConstraints strtStpCstrts = new GridBagConstraints();
			final JButton start = new JButton("Start");
			final JButton stop = new JButton("Stop");
			stop.setEnabled(false);

			start.addActionListener(new ActionListener() {

				@Override
				public void actionPerformed(ActionEvent e) {
					try {
						setupAlgorithm();
						start.setEnabled(false);
						stop.setEnabled(true);
						mLocTask = new LocalizerTask();
						mLocTask.execute();
					} catch (IOException ioe) {
						ioe.printStackTrace();
						SwingUtilities.invokeLater(new ShowErrorDialog(
								"I/O error", ioe));
					} catch (UnsupportedAudioFileException uafe) {
						uafe.printStackTrace();
						SwingUtilities.invokeLater(new ShowErrorDialog(
								"Audio file error", uafe));
					} catch (LineUnavailableException luae) {
						luae.printStackTrace();
						SwingUtilities.invokeLater(new ShowErrorDialog(
								"Unable to open audio", luae));
					} catch (Exception exc) {
						exc.printStackTrace();
						SwingUtilities.invokeLater(new ShowErrorDialog(
								"Error on algorithm setup", exc));
					}
				}
			});

			stop.addActionListener(new ActionListener() {

				@Override
				public void actionPerformed(ActionEvent e) {
					mLocTask.cancel(false);
					stop.setEnabled(false);
					start.setEnabled(true);
					mAzDisp.clear();
				}
			});

			strtStpCstrts.gridx = 0;
			strtStpCstrts.gridy = 0;
			strtStpCstrts.weightx = 0.5;
			strtStpCstrts.fill = GridBagConstraints.HORIZONTAL;
			add(start, strtStpCstrts);
			strtStpCstrts.gridx = 1;
			strtStpCstrts.gridy = 0;
			strtStpCstrts.weightx = 0.5;
			strtStpCstrts.fill = GridBagConstraints.HORIZONTAL;
			add(stop, strtStpCstrts);
		}
	}

	class LocalizerTask extends
			SwingWorker<Void, List<LocalizationAlgorithm.Az>> {
		@SuppressWarnings("unchecked")
		@Override
		public Void doInBackground() {
			try {
				while (!isCancelled()) {
					publish(mAlg.compute());
				}
			} catch (IOException e) {
				e.printStackTrace();
				SwingUtilities
						.invokeLater(new SoundLocalizerWindow.ShowErrorDialog(
								"Audio read error", e));
			}
			return null;
		}

		@Override
		protected void process(List<List<LocalizationAlgorithm.Az>> a) {
			if (!isCancelled() && (a.size() > 0)) {
				List<LocalizationAlgorithm.Az> peaks = a.get(a.size() - 1);
				mAzDisp.updatePeaks(peaks);
			}
		}
	};

	private static class ShowErrorDialog implements Runnable {
		private String msg;

		public ShowErrorDialog(String msg, Exception e) {
			this.msg = "<html><b>" + msg + ":</b>";
			this.msg += "<p>" + e.getMessage() + "</p>";
		}

		@Override
		public void run() {
			JOptionPane.showMessageDialog(null, msg, "Error",
					JOptionPane.ERROR_MESSAGE);

		}
	}
}
