/* SoundLocalizer - a sound source localization program
 * Copyright (C) 2010 Laurent Calmes
 *
 * This file is part of SoundLocalizer.
 *
 * SoundLocalizer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SoundLocalizer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SoundLocalizer.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.laurentcalmes.soundloc;

import java.io.IOException;

import javax.sound.sampled.AudioFormat;

/**
 * Implements an <tt>AudioDevice</tt> which can synthesize three different
 * signal types with an interaural time difference corresponding to a specified
 * azimuth. Note that the time difference is realized through a sample shift.
 * This means that the ITD - and hence the azimuth - can only be approximated.
 * The higher the sampling frequency, the higher the accuracy.
 * 
 * @author Laurent Calmes
 * 
 */
public class SynthDevice extends AudioDevice {
	private AudioFormat mAf;
	private int mVirtITDspls;
	private double mVirtITDsecs;
	private SigType mType;
	private long mAmpOffset;
	private long mAmpMax;
	private float mSineFreq;
	private long mSineSplNum = 0;

	/**
	 * Available signal types for <tt>SynthDevice</tt>
	 * 
	 * @author Laurent Calmes
	 *
	 */
	public static enum SigType {
		IMPULSE, NOISE, SINE
	}

	/**
	 * @param type
	 *            desired signal type
	 * @param virtAz
	 *            desired azimuth to the virtual sound source
	 * @param baseline
	 *            microphone distance in metres
	 * @param tfSize
	 *            Timeframe size in seconds
	 * @param af
	 *            Desired <tt>AudioFormat</tt>
	 */
	public SynthDevice(SigType type, double virtAz, double baseline,
			double tfSize, AudioFormat af) {
		super(tfSize);
		mType = type;
		mAf = new AudioFormat(af.getSampleRate(), af.getSampleSizeInBits(), 2,
				true, false);
		mAmpMax = 1 << af.getSampleSizeInBits();
		mAmpOffset = 1 << (af.getSampleSizeInBits() - 1);
		double maxITDsecs = baseline / LocalizationAlgorithm.soundSpeed;
		mVirtITDsecs = Math.sin(virtAz * Math.PI / 180.0) * baseline
				/ LocalizationAlgorithm.soundSpeed;
		if (Math.abs(mVirtITDsecs) > maxITDsecs) {
			mVirtITDsecs = maxITDsecs * Math.signum(mVirtITDsecs);
		}
		mVirtITDspls = (int) Math.round(mVirtITDsecs * af.getSampleRate());
	}

	/**
	 * Sets the frequency for the sine signal type
	 * 
	 * @param freq
	 */
	public void setSineFreq(float freq) {
		mSineFreq = freq;
	}

	@Override
	public AudioFormat getFormat() {
		return mAf;
	}

	@Override
	protected void readRaw(byte[] rawData) throws IOException {
		long[] left = new long[mTFSizeSpls];
		long[] right = new long[mTFSizeSpls];
		switch (mType) {
		case IMPULSE:
			impulse(left, right);
			break;
		case NOISE:
			noise(left, right);
			break;
		case SINE:
			sine(left, right);
			break;
		default:
			throw new IOException(
					"SynthDevice.readRaw(): Unknown signal type selected!");
		}

		for (int i = 0, j = 0; i < mTFSizeSpls; ++i) {
			long tmp = left[i];
			for (int k = 0; k < mSampleSizeInBytes; ++k) {
				rawData[j++] = (byte) (tmp & 255);
				tmp >>= 8;
			}
			tmp = right[i];
			for (int k = 0; k < mSampleSizeInBytes; ++k) {
				rawData[j++] = (byte) (tmp & 255);
				tmp >>= 8;
			}
		}
	}

	private void impulse(long[] left, long[] right) {
		left[mTFSizeSpls >> 1] = right[(mTFSizeSpls >> 1) - mVirtITDspls] = mAmpMax
				- mAmpOffset;
	}

	private void noise(long[] left, long[] right) {
		int sampleDiffL, sampleDiffR;
		sampleDiffR = mVirtITDspls / 2;
		if ((mVirtITDspls % 2) != 0) {
			if (mVirtITDspls < 0)
				sampleDiffL = -mVirtITDspls / 2 + 1;
			else
				sampleDiffL = -mVirtITDspls / 2 - 1;
		} else {
			sampleDiffL = -mVirtITDspls / 2;
		}
		int offset = (Math.abs(sampleDiffL) > Math.abs(sampleDiffR) ? Math
				.abs(sampleDiffL) : Math.abs(sampleDiffR));
		long[] noise = new long[2 * offset + mTFSizeSpls];
		for (int i = 0; i < mTFSizeSpls; ++i) {
			noise[i] = Math.round(mAmpMax * Math.random()) - mAmpOffset;
		}

		for (int i = 0; i < mTFSizeSpls; ++i) {
			left[i] = noise[offset + sampleDiffL + i];
			right[i] = noise[offset + sampleDiffR + i];
		}
	}

	private void sine(long[] left, long[] right) {
		double phaseDelay_2 = mVirtITDsecs * Math.PI * mSineFreq;

		for (int i = 0; i < mTFSizeSpls; ++i) {
			double t = mSineSplNum++ / mAf.getSampleRate();
			left[i] = Math.round(mAmpMax
					* Math.sin(2.0 * Math.PI * mSineFreq * t - phaseDelay_2))
					- mAmpOffset;
			right[i] = Math.round(mAmpMax
					* Math.sin(2.0 * Math.PI * mSineFreq * t + phaseDelay_2))
					- mAmpOffset;
		}
	}
}
